<?php

class WizardBehavior extends CBehavior {
	const ERR_ACTION_NOT_FOUND = 0;
	const ERR_SESSION_LIMIT_REACHED = 1;
	const ERR_WID_NOT_FOUND = 2;
	const ERR_WID_EXPIRED = 3;
	
	
	/**
	 * action array in form of array('action',array('title'=>'actionName',...))
	 * @var array
	 */
	public $actions;
	
	/**
	 * 
	 * @var int
	 */
	public $errorCode=0;
	
	/**
	 * 
	 * @var mixed
	 */
	public $errorMsg='';
	
	/**
	 * maximum wizardSessions per user
	 */
	public $maxSessions = 50;
	
	/**
	 * @var string
	 */
	public $priorities=array('param','page');
	
	/**
	 * @var string
	 */
	public $queryParam='ywz';
	
	/**
	 * @var string
	 */
	public $sessionKeyPrefix='__';
	
	/**
	 * wizard expires after this ammount of seconds
	 * @var int
	 */
	public $sessionTimeout=900;
	
	/**
	 * <p>the string used as a variable for the current step in URL. Its only 
	 * used if this behavior is used with an action.</p>
	 * @var string
	 */
	public $stepParam='step';
	
	/**
	 * the id to discriminate this wizard from others wizards used over the application
	 * @var string
	 */
	public $uniqueId;
	
	/**
	 * whether this wizard shall handover the wizardId visible or hidden to the end user
	 * @var boolean
	 */
	public $visibleMode=true;
	
	/**
	 * @param CComponent $owner
	 */
	public function attach($owner){
		parent::attach($owner);
		$this->attachHandlers($owner);
		$this->init();
	}
	
	private function attachHandlers($owner) {
		$this->attachEventHandler('onError',array($owner,'errorHandler'));
		$this->attachEventHandler('onValidate',array($owner,'validateHandler'));
        $this->attachEventHandler('onStart',array($owner,'startHandler'));
        $this->attachEventHandler('onNavigate',array($owner,'navigationHandler'));
        $this->attachEventHandler('onFinish',array($owner,'finishHandler'));
	}

	/**
	 * <p>this method removes any data set for the provided $sessionKey. If none is delivered it uses 
	 * the session key of the current wizard instance</p>
	 * @return void
	 */
	public function clearData($sessionKey=null){
		if(null===$sessionKey)
			$sessionKey=$this->getSessionKey();

		//remove wizard data from session
		$session=$this->getSession();
		foreach($this->_keys as $key)
			unset($session[str_replace('{sessionKey}',$sessionKey,$key)]);

		//remove sessionKey from keyList
		$this->removeKey($sessionKey);
		
		return true;
	}
	
	public function clear() {
		$this->clearData();
	}
	
	/**
	 * checks if a session key is expired
	 * @param string $sessionKey optional, if not set evaluation is made retrieving the sessionKey of this wizard instance
	 * @return boolean true if session is expired, false otherwise
	 */
	public function expired($sessionKey=null){
		if(null===$sessionKey)
			$sessionKey=$this->getSessionKey();
		
		$expired = false;
		//if sessionKey is not set or has timeout
		$expired = (isset($session[$sessionKey]) && time()>$session[$sessionKey.'timeout']);
		//if wizardId is not set in keyList
		return ($expired || !$this->getKeyList()->contains($sessionKey));

//		$session=$this->getSession();
//		return (isset($session[$sessionKey]) && time()>$session[$sessionKey.'timeout']);
		//if $sessionKey is not set in keyMap it has never been existed or is expired. Interprete both as expired in this context.
//		return $expired;
	}
	
	/**
	 * checks whether an action has already been visited
	 * @param string $action optional, if not set the evaluation is based on getCurrentAction()
	 */
	public function isActionVisited($action=null){
		if(null===$action)
			$action=$this->getCurrentAction();

		return in_array($action,$this->getVisitedActions()->toArray(),true);
	}
	
	/*
	 **************************
	 ** GETTER UND SETTER **
	 **************************
	 */	
	/**
	 * Tries to return an action for a certain step. If no step parameter is set, it return the curent action
	 * @param int $step
	 * @param int $default
	 * @return the action or $default string if no such action exists.
	 */
	public function getAction($step=-1,$default=''){
		$step = (0>$step)?  $this->getStep() : $step;
		if($this->_routes->offsetExists($step))
			return $this->_routes->offsetGet($step);
		
		$this->error(WizardBehavior::ERR_ACTION_NOT_FOUND,array('action'=>'','step'=>$step,'offet'=>0));
		return $default;
	}
	
	/**
	 * returns the url for a certain action.
	 * @param int $offset
	 * @param array $params
	 * @param string $ampersand
	 * @return string the url for $action
	 */
	public function getActionUrl($action,$params=array(),$ampersand='&') {
		return $this->getStepUrl($this->getStep($action),$params,$ampersand);
	}
	
	/**
	 * returns the current action based on an action->route <-> currentRoute comparation 
	 * @return string returns action or empty string if no action is found
	 */
	public function getCurrentAction() {
		
		if(null!==$this->_currentAction)
			return $this->_currentAction;
		
		if ($this->visibleMode) {
			//retrieve action string from params, if attached to CAction
			if($this->_isActionBehavior) {
				$action=Yii::app()->getRequest()->getParam($this->stepParam,'');
				$action=(''===$action)? $this->getAction(0) : $action;
			//retrieve action string from url, if attached to CController
			} else {
				$action = $this->getCurrentRoute();
				$action = substr($action,strrpos($action,'/')+1);
			}

		} else {
			/*
			 * If page state is set return appropriate action. Otherwise this wizard
			 * is handled as not yet initialized and defaults to first action.
			 */
			$action = $this->getAction($this->getPageState($this->stepParam,0));		
		}
		
		//setup result of action query 
		if(-1!==($i=$this->_routes->indexOf($action))) {
			$this->_currentAction = $this->_routes->itemAt($i);
		} else {
			$this->_currentAction = '';
			$this->error(WizardBehavior::ERR_ACTION_NOT_FOUND,array('action'=>$action,'step'=>$i,'offet'=>0));
		}
			
		
		//if no appropriate action was found return empty string
		return $this->_currentAction;
	}
	
	/**
	 * returns the current wizard step
	 * @return int if available this method returns current step, otherwise -1
	 */
	public function getCurrentStep() {
		if(null!==$this->_currentStep)
			return $this->_currentStep;
			
		$this->_currentStep = $this->getStep($this->getCurrentAction());
		return $this->_currentStep;
	}

	/**
	 * returns the data already stored in this wizard as CMap.
	 * @return CMap
	 */
	public function getData(){
		$session=$this->getSession();
		$key=$this->getSessionKey();
		
		if (isset($session[$key]) && ($session[$key] instanceof CMap)) {
			return $session[$key];
		} else {
			return $session[$key]=new CMap();
//			$this->clearData($key);
//			$this->errorCode=404;
//			$this->errorMsg='SESSION_EXPIRED';
//			$this->error();
		}
		//run garbage collection
//		$this->gc();
//		if(isset($session[$key]) && ($session[$key] instanceof CMap))
//			return $session[$key];
//		else
//			return $session[$key]=new CMap();
	}

	/**
	 * this method returns the items in array notation as e.g usable as CMenu input
	 * @param $skipActions array of actions that shall be marked as invisible
	 * @param $widVisible decide whether wizardId should be appended to url parameters. Default is true. This option is only meaningful for invisible mode.
	 * @return array of contained items
	 */
	public function getItems(array $skipActions=array(),$widVisible=true){
		$params = array();
		if($widVisible) {
			$params = array($this->queryParam=>$this->getWizardId());
		}
			
		$this->_items = array();
		foreach($this->_routes->toArray() as $key=>$action) {
			$items[] = array(
				'label'=>$this->_titles[$key],
				'url'=>$this->getStepUrl($key,$params),
				'visible'=>!in_array($action,$skipActions),
				'active'=>($action==$this->getCurrentAction())
			);
		}
		return ($this->_items = $items);
	}
	
	/**
	 * get value for the variable with key "id"
	 * @param string $id
	 * @param mixed $defaultValue
	 * @return mixed
	 */
	public function getState($id,$defaultValue=null){
		$data=$this->getData();
		if(null===($value=$data[$id]))
			$value=$defaultValue;
		return $value;
	}
	
	/**
	 * set value for the variable with key "id". 
	 * @param string $id the key
	 * @param mixed $value use mixed type to add and null to remove a value.
	 */
	public function setState($id,$value){
		$data=$this->getData();
		($value===null)?
			$data->remove($id):
			$data->add($id,$value);
	}

	/**
	 * returns the step key for a certain action, defaults to current action
	 * @param string $action defaults to current action. If you want to use the offset variable, with current action, set it to empty string
	 * @param int $offset the index offset relative to the input action step 
	 * @param int $default the value this method returns if no step is found for the given input parameter
	 * @return int -1 if the input action or the step the offset points to don't exist, step number otherwise
	 */
	public function getStep($action='',$offset=0,$default=-1) {
		//if action not set default to current action
		$action = (''===$action)? $this->getCurrentAction() : $action;
		//if action is given as array, retrieve string elemente at key 0
		$action = (is_array($action))? $action[0] : $action;  
		//get index of action
		$index = (''!==$action && $this->_routes->contains($action) )? 
			$this->_routes->indexOf($action) : $default;
			
		//return index of action(+offset) or default if failed
		$step = (-1!==$index && (0===$offset || $this->_routes->offsetExists($index+$offset)) )? 
			$index+$offset : $default;
			
		if (-1===$step)
			$this->error(WizardBehavior::ERR_ACTION_NOT_FOUND,array('action'=>$action,'step'=>$step,'offet'=>$offset));

		return $step;
	}
	
	/**
	 * returns the url for a certain step
	 * @param $step
	 * @param $params
	 * @param $ampersand
	 * @return string the url for the $step
	 */
	public function getStepUrl($step,$params=array(),$ampersand='&'){
		//if visible mode is on, set wizardID as url parameter
		if ($this->visibleMode) {
			//interlace wizardId
			$params = array_merge(array($this->queryParam=>$this->getWizardId()),$params);
			//merge provided parameter with internal parameters
			$params = array_merge($this->_params[$step],$params);
		}
		//finally return aggregated url
		return Yii::app()->createUrl($this->_urls[$step],$params);
	}
	
	/**
	 * returns the title for a certain step, defaults to current step
	 * @param integer $n step
	 * @param string $default the default value 
	 * @return string if set it returns the title, otherwise $default
	 */
	public function getTitle($step=-1,$default=''){
		$step = (is_int($step) && $step>=0)? $step : (int)$this->getCurrentStep();
		if(isset($this->_titles[$step]))
			return $this->_titles[$step];
		return '';
	}
	
	/**
	 * this method returns an CList of all already visited actions of this wizard
	 * @return CList
	 */
	public function getVisitedActions(){
		$session=$this->getSession();
		$key=$this->getSessionKey().'actions';

		if(isset($session[$key]) || ($session[$key] instanceof CList))
			return $session[$key];
		else
			return $session[$key]=new CList();
	}
	
	/**
	 * <p>Retrieves the current wizard id. If not already set it will create a new one.</p>
	 * @return <b>string</b> This wizards Id
	 */
	public function getWizardId(){
		return $this->wizardId;
//		if(null===$this->wizardId){
//			//retrieve the query param from GET or POST
//			$id=Yii::app()->request->getParam($this->queryParam);
//			if(empty($id) || $this->expired($this->sessionKeyPrefix.$id)) {
//			
//				//retrieve latest session key,  e.g. default $sessionKey = __ywz1 - __ywz1# 
//				$sessionKey = $this->sessionKeyPrefix.$this->queryParam.self::$_counter;
//				
//				// find first sessionKey that is neither in use nor expired
//				for($session=$this->getSession(); isset($session[$sessionKey]) || $this->expired($sessionKey); /* DO NOTHING */ ) {
//					//ensure that $counter does not become greate than maxSessions
//					if(++self::$_counter > ~-$this->maxSessions)
//						throw new CHtmlException(413,'Too many wizard creation attempts, try again in '.$this->sessionTimeout.' seconds');
//					$sessionKey = $this->sessionKeyPrefix.$this->queryParam.self::$_counter;
//				}
//				// set free id from for loop as this wizards id
//				$this->wizardId=$this->queryParam.(self::$_counter);
//				
//			} else {
//				$this->wizardId=Yii::app()->request->getParam($this->queryParam);
//			}
//		}
//		return $this->wizardId;
	}
	
	/**
	 * sets the wizard id to the delivered value
	 * @param string $value
	 */
	public function setWizardId($value) {
		$this->wizardId=$value;
	}
	
	/**
	 * whether this wizard is already connected with an id 
	 * @return boolean
	 */
	public function hasId() {
		return(null!==$this->wizardId && !empty($this->wizardId));
	}
	
	public function redirect($step) {
		
		if (!is_int($step) || !$this->_routes->offsetExists($step)) {
			$this->errorCode = WizardBehavior::ERR_STEP_NOT_FOUND;
			$this->error();
		} else {
			if ($this->visibleMode) {
				$c = ($this->_isActionBehavior)? $this->getOwner()->getController() : $this->getOwner();
				$c->redirect($this->getStepUrl($step));
			} else {
				//simulate steps for 'real' redirect
				$_POST = array();
				$this->setPageState($this->queryParam,$this->getWizardId());
				$this->setPageState($this->stepParam,$step);
				$this->_currentAction=$this->getAction($step);
				$this->_currentStep=$step;
				$this->processStep();
				Yii::app()->end();
//				$this->init();
//				$this->runWizard();
			}
		}
	}
	
	/**
	 * This method runs this wizard beginning with the last
	 * saved state.
	 * @param <b>$autoId</b> Set if wizard should try to autoload its wizardId
	 * @return void
	 */
	public function runWizard() {
		//raise validation event
		$this->validate();
		$this->processStep();
	}
	
	private function processStep() {
		$n=$this->getCurrentStep();
		//mark action as visited
		$this->visitAction($n);
		// run events
		if(-1===$n) {
			$this->errorCode = WizardBehavior::ERR_STEP_NOT_FOUND;
			$this->error();
		} else if(0===$n)
			$this->start();
		else if($n>=~-count($this->actions))
			$this->finish();
		else if(is_int($n))
			$this->navigate();
	}
	
	/*
	 **************************
	 ** EVENT SOURCE METHODS **
	 **************************
	 */
	/**
	 * raises an onFinish event
	 * @param CEvent $event
	 * @return void
	 */
	public function onError($event){
		$this->raiseEvent('onError',$event);
	}
	

	/**
	 * raises an onFinish event
	 * @param CEvent $event
	 * @return void
	 */
	public function onFinish($event){
		$this->raiseEvent('onFinish',$event);
	}
	
	/**
	 * raises an onNavigate
	 * @param CEvent $event
	 * @return void
	 */
	public function onNavigate($event){
		$this->raiseEvent('onNavigate',$event);
	}
	
	/**
	 * raises an onStart event
	 * @param CEvent $event
	 * @return void
	 */
	public function onStart($event){
		$this->raiseEvent('onStart',$event);
	}

	/**
	 * raises an onValidate event
	 * @param CEvent $event
	 * @return void
	 */
	public function onValidate($event){
		$this->raiseEvent('onValidate',$event);
	}
	
	/*
	 **************************************
	 ** PROTECTED ATTRIBUTES AND METHODS **
	 **************************************
	 */	
	
	/**
	 * The Id of this wizard
	 * @var string
	 */
	protected $wizardId;
	
	/**
	 * creates an onError event and sends it to onError() event raiser.
	 * @return unknown_type
	 */
	protected function error($code,$msg=null) {
		$this->errorCode=$code;
		$this->errorMsg=$msg;
		if($this->hasEventHandler('onError'))
			$this->onError(new CEvent($this));	
	}

	/**
	 * creates an onStart event and sends it to onFinish() event raiser.
	 * @return void
	 */
	protected function finish(){
		if($this->hasEventHandler('onFinish'))
			$this->onFinish(new CEvent($this));
		Yii::app()->onEndRequest[]=array($this,'clear');
	}
	
	/**
	 * creates an onStart event and sends it to onNavigate() event raiser.
	 * @return void
	 */
	protected function navigate(){
		if($this->hasEventHandler('onNavigate'))
			$this->onNavigate(new CEvent($this));
	}
	
	/**
	 * creates an onStart event and sends it to onStart() event raiser.
	 * @return void
	 */
	protected function start(){
		if($this->hasEventHandler('onStart'))
			$this->onStart(new CEvent($this));
	}
	
	/**
	 * creates an onValidate event and sends it to onValidate() event raiser.
	 * @return void
	 */
	protected function validate(){
		if($this->hasEventHandler('onValidate'))
			$this->onValidate(new CEvent($this));
	}
	
	/**
	 * this method marks the current action as visited
	 * @param string $action
	 */
	protected function visitAction($step) {
		$action=$this->getAction($step);

		if(!$this->isActionVisited($action))
			$this->getVisitedActions()->add($action);
	}

	/*
	 ************************************
	 ** PRIVATE ATTRIBUTES AND METHODS **
	 ************************************
	 */
	
	/**
	 * internal counter used for retrieval of new wizard ids.
	 * @var int
	 */
	static private $_counter=0;
	
	/**
	 * buffer for current action.
	 * @var string
	 */
	private $_currentAction;
	private $_currentStep;
	
	/**
	 * determines if this behavior is attached to an CAction (<b>true</b>)
	 * or to an CController (<b>false</b>).
	 * @var boolean
	 */
	private $_isActionBehavior;
	
	/**
	 * @var array
	 */
	private $_keys=array(
		'{sessionKey}',
		'{sessionKey}timeout',
		'{sessionKey}actions',
	);
	
	/**
	 * The map holding the wizard 
	 */
	private $_keymap = array();
	
	/**
	 * buffer for a menu representing the stored actions
	 * @var CMenu
	 */
	private $_items;
	
	/**
	 * The params for a certain action
	 * @var array
	 */
	private $_params = array();

	/**
	 * The action store in CList format 
	 * @var CList
	 */
	private $_routes;
	
	/**
	 * The action titles for displaying
	 * @var array
	 */
	private $_titles = array();
	
	/**
	 * The base url for this behavior e.g. site/ or site/register
	 * @var string
	 */
	private $_urlBase = '';
	
	/**
	 * The urls for a certain action
	 * @var array
	 */
	private $_urls = array();
	
	/**
	 * this method creates and returns an URL for an action
	 * @param string $action
	 * @param array $params
	 * @param string $ampersand
	 * @return string
	 */
//	private function createUrl($step,$params=array(),$ampersand='&'){
//		$params = array_merge($params, $this->_params[$step]);
//		return Yii::app()->createUrl($this->_urls[$step],$params);
//	}

	/**
	 * creates a random key for new wizard
	 * @return <b>string</b> a random key
	 */
	private function createRandomKey() {
		srand((double)microtime()*1000000);
        $rand64="./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";
        $key = '';
        for ($index = 0; $index < 4; $index++) {
            $random=rand();
            $key.=substr($rand64,$random%64,1).substr($rand64,($random/64)%64,1);
        }
        $key=substr($key,0,8);
        return($key);
	}

	private function createWizardId() {
		$randomKey=$this->createRandomKey();
		while (-1===$this->addKey($this->sessionKeyPrefix.$randomKey)) {
			$randomKey=$this->createRandomKey();
		}
		$this->wizardId=$randomKey;
//		//retrieve latest session key,  e.g. default $sessionKey = __ywz1 - __ywz1# 
//		$sessionKey = $this->sessionKeyPrefix.$this->queryParam.self::$_counter;
//		
//		// find first sessionKey that is neither in use nor expired
//		for($session=$this->getSession(); isset($session[$sessionKey]) || $this->expired($sessionKey); /* DO NOTHING */ ) {
//			//ensure that $counter does not become greate than maxSessions
//			if(++self::$_counter > ~-$this->maxSessions)
//				throw new CHtmlException(413,'Too many wizard creation attempts, try again in '.$this->sessionTimeout.' seconds');
//			$sessionKey = $this->sessionKeyPrefix.$this->queryParam.self::$_counter;
//		}
//		// set free id from for loop as this wizards id
//		$this->wizardId=$this->queryParam.(self::$_counter);
	}
	
	/**
	 * the garbage collector method cleans up the values saved in expired sessions
	 * @return void
	 */
	private function gc(){
//		$session=$this->getSession();
//		$keymap = $this->getKeyMap();
		foreach ($this->getKeyList()->toArray() as $sessionKey) {
			if ($this->expired($sessionKey))
				$this->clearData($sessionKey);
		}
//		$session[$this->getSessionKey().'timeout']=time()+$this->sessionTimeout;
		
		/*
		 * clear data for expired sessions at least from 0..$this->maxSessions. If $i++ finds sessions > $this->maxSessions 
		 * it will remove them as long as they occur in a row
		 */ 
//		for($items=$this->maxSessions,$i=0;
//			($this->expired($sessionKey=$this->sessionKeyPrefix.$this->queryParam.($i++)) && $this->clearData($sessionKey)) || $items--;);
//		for($items=$this->maxSessions,$i=0;
//			($this->expired($sessionKey=$this->sessionKeyPrefix.$this->queryParam.($i++)) && $this->clearData($sessionKey)) || $items--;);
	}
	
	/**
	 * returns the current route for this wizard. If not set this method tries to determine the current 
	 * route via the CController->getRoute() and if this fails via the UrlManager->parseUrl() and returns
	 * the result.
	 * @return string the current route
	 */
	private function getCurrentRoute() {
		if(null!==($controller=Yii::app()->getController())) {
			return $controller->getRoute();
		} else {
			return Yii::app()->getUrlManager()->parseUrl(Yii::app()->getRequest());			
		}
	}
	
	private function getPageState($key,$value) {
		$c = ($this->_isActionBehavior)? $this->getOwner()->getController() : $this->getOwner();
		return $c->getPageState($key,$value);
	}
	
	private function setPageState($key,$value) {
		$c = ($this->_isActionBehavior)? $this->getOwner()->getController() : $this->getOwner();
		$c->setPageState($key,$value);
	}
	
	/**
	 * retrieve session for the current Webapp user
	 * @return CHttpSession
	 */
	private function getSession(){
		return Yii::app()->getComponent('session');
	}
	
	/**
	 * returns the current wizards session key
	 * @return string
	 */
	private function getSessionKey(){
		return $this->sessionKeyPrefix.$this->getWizardId();
	}
	
	/**
	 * initializes this wizards behavior
	 * @return void
	 */
	private function init(){

		// checks if owner is present
		$owner=$this->getOwner();
		if($owner===null) {
			throw new CException('no owner is set');
		}
		
		// checks if owner has id
		if (empty($owner->id)) {
			throw new CException('owner has no id');
		}
		
		// checks if $this is attached to controller or action
		if (is_a($owner,'CAction')) {
			$this->_isActionBehavior = true;			
		} elseif(is_a($owner,'CController')) {
			$this->_isActionBehavior = false;
		} else {
			throw new CException('This behavior is only intended to use with CAction or CController');
		}
		
		/*
		 ************************************
		 ** setup actions, urls and params **
		 ************************************
		 */
		//construct base path for routes
		$base = (isset($owner->module,$owner->module->id))? $owner->module->id.'/' : '';
			if($this->_isActionBehavior) {
			$base .= $owner->getController()->id.'/';
			$base .= $owner->id;			
		} else {
			$base .= $owner->id.'/';
		}

		//prepare action list and arrays for urls, url parameters, and action titles
		$this->_routes = new CList();
		foreach(((array)$this->actions) as $action) {
			$key = $this->_routes->add($action[0]);
			$this->_params[$key] = ($this->_isActionBehavior)? array($this->stepParam=>$action[0]) : array();
			$this->_urls[$key] = ($this->_isActionBehavior)? $base : $base.$action;
			$this->_titles[$key] = (isset($action['title']))? $action['title'] : '';
		}
		//make list read only
		//TODO Bug report, read only list should not be able to get new data copied
		$this->_routes = new CList($this->_routes,true);
		
		$this->_currentAction = $this->getCurrentAction();
		$this->_currentStep = $this->getCurrentStep();
		
//		$this->gc();
		
		if (0===$this->_currentStep) {
			$this->createWizardId();
		} else {
			$wid=$this->loadWizardId();
			if (empty($wid)) {
				$this->error(WizardBehavior::ERR_WID_NOT_FOUND);
			}
			else if ($this->expired($this->sessionKeyPrefix.$wid)) {
				$this->error(WizardBehavior::ERR_WID_EXPIRED,$wid);
			} else {
				$this->setWizardId($wid);
			}
		}
		
		//Save WizardId in Page State
		if (!$this->visibleMode) {
			$this->setPageState($this->queryParam,$this->getWizardId());
			$this->setPageState($this->stepParam,$this->getCurrentStep());
		}
		
	}
	
    /**
     * This method tries to connect this wizard behavior with an wizardId stored as 
     * GET/POST parameter, PageState or in a SESSION variable.
     * @param <b>$priorities</b> an array that may consist of the values 'param', 
     * 'page' and 'session'. It tries to retrieve the wizard Id in the given order 
     * and only for the method pointers contained in priorities. It defaults to 
     * array('param','page','session').
     * @return <b>string</b> WizardID if successfull, empty string otherwise
     */
    private function loadWizardId() {
    	foreach($this->priorities as $p) {
    		switch($p) {
    			case 'param':
			    	//Trying to retrieve wizardId from GET/POST param
					if (null!==($wid=Yii::app()->getRequest()->getParam($this->queryParam))) {
			        	$this->setWizardId($wid);
			        	return $wid;
					}
					break 1;
					
    			case 'page':
		    		//Trying to retrieve wizardId from Page state
			    	$c = ($this->_isActionBehavior)? $this->getOwner()->getController() : $this->getOwner(); 
					if (null!==($wid=$c->getPageState($this->queryParam))) {
			        	$this->setWizardId($wid);
			        	return $wid;
					}
					break 1;
					
    			case 'session':
    				//Trying to retrieve wizardId from user session
				 	if (Yii::app()->user->hasState($this->queryParam) 
			        		&&  null!==($wid=Yii::app()->user->getState($this->queryParam))) {
			        	$this->setWizardId($wid);
			        	return $wid;
			        }
			        break 1;
			        
			     default:
			     	throw new CException('retrieval method \''.$p.'\' does not exist');
    		}
    	}
        //retrieval failed
        return null;
    }
    
    private function getKeyList() {
    	$session=$this->getSession();
    	$key=$this->uniqueId.'keymap';
    	if ($session->contains($key)) {
    		return $session[$key];
    	} else {
    		return $session[$key]=new CList;
    	}
    }
    
    private function addKey($value) {
    	$list=$this->getKeyList();
    	//start garbage collection of key list is almost full
    	if($keyList->count > ceil(0.8*$this->maxSessions)) {
			$this->gc();
			$keyList = $this->getKeyList();
		}
		//raise error if maxSessions exceede
    	if($list > ~-$this->maxSessions) {
    		$this->errorCode=404;
    		$this->errorMsg='MAX_WIZARD_SESSIONS_EXCEEDED';
    		$this->error();
    	} else {
    		if($list->contains($value)) {
    			return -1;
    		} else {
    			return $list->add($value);
    		}
    	}
    }
    
    private function removeKey($value) {
    	$list=$this->getKeyList();
    	if(!$list->contains($value)) {
   			return -1;
   		} else {
   			return $list->remove($value);
   		}
    }
}